@@ -114,14 +114,15 @@ static gint m_GeometryY = -1;
 static Gok_Predictor m_pCommandPredictor = NULL;
 
 /* Private functions */
-gint gok_main_open(gint argc, gchar *argv[]);
-void gok_main_close (void);
-void gok_main_read_rc (void);
-void gok_main_initialize_access_methods (void);
-void gok_main_initialize_wordcomplete (void);
-void gok_main_initialize_commandprediction (void);
-gboolean gok_main_display_scan_main(void);
-void gok_main_display_geometry_error (void);
+static void gok_main_read_rc (void);
+static void gok_main_initialize_access_methods (void);
+static void gok_main_initialize_wordcomplete (void);
+static void gok_main_initialize_commandprediction (void);
+static gboolean gok_main_display_scan_main(void);
+static void gok_main_display_geometry_error (void);
+static void gok_main_object_state_listener (Accessible* pAccessible);
+static gboolean gok_main_has_xkb_extension ();
+static gboolean gok_main_check_sticky_keys (GtkWidget *widget);
 
 /* poptOption storage */
 static char* accessmethodname = NULL;
@@ -293,7 +294,8 @@ gint main (gint argc, gchar *argv[])
 *
 * returns: a program exit code.
 **/
-gint gok_main_open(gint argc, gchar *argv[])
+gint
+gok_main_open(gint argc, gchar *argv[])
 {
 	GnomeProgram *gok_program;
 	/* initialize member data */
@@ -354,6 +356,11 @@ gint gok_main_open(gint argc, gchar *argv[])
 	/* read the user profile data */
 	gok_data_initialize ();
 
+	/* initialize the word prediction engine with gconf values */
+	if (gok_data_get_use_aux_dictionaries ())
+		gok_wordcomplete_set_aux_dictionaries (gok_wordcomplete_get_default (),
+						       gok_data_get_aux_dictionaries ());
+
 	gok_log ("gok_data_initialize has finished");
 
 	/* load the ".rc" files used for the key styles */
@@ -414,7 +421,8 @@ gint gok_main_open(gint argc, gchar *argv[])
 	gok_spy_open();
 	gok_spy_register_windowchangelistener ((void *)gok_main_window_change_listener);
 	gok_spy_register_mousebuttonlistener ((void *)gok_main_mousebutton_listener);	
-
+	gok_spy_register_objectstatelistener ((void *)gok_main_object_state_listener);
+	
 	/* initialize sound */
 	gok_sound_initialize();
 
@@ -486,7 +494,7 @@ gok_log ("check for currently active frame");
 gok_log ("finished check");
 	
 	/* connect signal handlers; don't trap SIGKILL */ 
-	signal(SIGSEGV, (void (*)(int))gok_sig_handler);
+	signal(SIGSEGV, (void (*)(int))gok_sig_handler); 
 	signal(SIGTERM, (void (*)(int))gok_sig_handler);
 	signal(SIGINT,  (void (*)(int))gok_sig_handler);
 
@@ -494,7 +502,8 @@ gok_log ("finished check");
 }
 
 /* private */
-gboolean gok_main_display_scan_main()
+static gboolean
+gok_main_display_scan_main()
 {
 	if (gok_main_display_scan ((GokKeyboard*)NULL, mainkeyboardname, KEYBOARD_TYPE_UNSPECIFIED, KEYBOARD_LAYOUT_NORMAL, KEYBOARD_SHAPE_BEST) == FALSE)
 	{
@@ -646,6 +655,48 @@ void gok_main_window_change_listener (Accessible* pAccessible)
 	gok_log_leave();	
 }
 
+/**
+* gok_main_object_state_listener
+* @pAccessible: Pointer to the foreground window's accessible interface.
+* 
+* This function is called each time the foreground window changes.
+*
+*/
+static void 
+gok_main_object_state_listener (Accessible* pAccessible)
+{
+	GokKeyboard* pKB;
+	GokKey* pK;
+	AccessibleStateSet* ass = NULL;
+
+	pKB = gok_main_get_current_keyboard();
+	if (pKB != NULL) {
+		pK = pKB->pKeyFirst;
+		while (pK != NULL) {
+			if (pK->pNodeAccessible != NULL) {
+				if (pK->pNodeAccessible->paccessible == pAccessible) {
+					/* note: this is very specific for now because we are doing
+					   this to make sure checkbox and radio menu item state
+					   is represented properly. TODO: generalize
+					*/
+					ass = Accessible_getStateSet(pAccessible);
+					if (AccessibleStateSet_contains( ass, SPI_STATE_CHECKED )) {
+						pK->ComponentState.active = TRUE;
+					}
+					else {
+						pK->ComponentState.active = FALSE;
+					}
+					AccessibleStateSet_unref(ass);
+					gok_key_update_toggle_state (pK);
+ 					break;
+				}
+			}
+			pK = pK->pKeyNext;
+		}
+	}
+}	
+
+
 /**
 * gok_main_motion_listener
 * @n_axes: Number of axes on which motion may have occurred
@@ -874,9 +925,8 @@ gboolean gok_main_display_scan_reset ()
 * gok_main_display_scan_previous_premade
 * 
 * Display and scan the first keyboard pulled from the branch-back-stack
-* that is premade. Delete any dynamic keyboards that are on the stack before
-* the first premade keyboard.
-* Don't push the current keyboard on the branch-back-stack.
+* that is premade. Delete any dynamic keyboards that are on the stack 'along
+* the way'. Don't push the current keyboard on the branch-back-stack.
 *
 * returns: TRUE if the previous keyboard was displayed, FALSE if not.
 */
@@ -896,24 +946,15 @@ gboolean gok_main_display_scan_previous_premade ()
 		return gok_main_display_scan_previous();
 	}
 
-	/* check this flag before branching back */
-/*
-	if (gok_data_get_dynamic_branchback() == FALSE)
-	{
-		return;
-	}
-*/
-	
 	/* initialize this array */
 	for (index = 0; index < MAX_DELETABLE_KEYBOARDS; index++)
 	{
 		arrayKeyboardsToDelete[index] = NULL;
 	}
-
+	
 	if (gok_branchbackstack_is_empty() == TRUE)
 	{
 		/* no keyboard on the branch back stack */
-		gok_log("no dynamic keyboards to remove during branch");
 		gok_log_leave();
 		return FALSE;
 	}
@@ -984,8 +1025,7 @@ gboolean gok_main_display_scan_previous ()
 	bReturnCode = TRUE;
 	pKeyboardToDelete = NULL;
 
-	if (gok_branchbackstack_is_empty() == TRUE)
-	{
+	if (gok_branchbackstack_is_empty() == TRUE) {
 		/* no keyboard on the branch back stack */
 		gok_log_leave();
 		return FALSE;
@@ -1005,14 +1045,11 @@ gboolean gok_main_display_scan_previous ()
 	{
 		/* delete the dynamically created keyboard */
 		/* make sure we branched back before deleting it */
-		if (bReturnCode == TRUE)
-		{
-			if (m_pCurrentKeyboard != pKeyboardToDelete) /* hack to fix potentially nastiness */
-			{
+		if (bReturnCode == TRUE) {
+			if (m_pCurrentKeyboard != pKeyboardToDelete) {
 				gok_keyboard_delete (pKeyboardToDelete,FALSE);
 			}
-			else
-			{
+			else {
 				gok_log_x("an attempt was made to delete the currently displayed keyboard!");
 			}
 		}
@@ -1155,7 +1192,8 @@ gboolean gok_main_ds (GokKeyboard* pKeyboard)
 	
 	/* layout the keys on the keyboard */
 	if (gok_keyboard_layout (pKeyboard, KEYBOARD_LAYOUT_NORMAL, 
-							KEYBOARD_SHAPE_SQUARE, FALSE) == FALSE)
+		 	/*KEYBOARD_SHAPE_WIDE KEYBOARD_SHAPE_SQUARE*/ 
+			KEYBOARD_SHAPE_KEYSQUARE, FALSE) == FALSE)
 	{
 		gok_log_leave();
 		return FALSE;
@@ -1267,7 +1305,8 @@ GtkWidget* gok_main_create_window (gboolean is_dock)
 * Retreives the name of the directory containing gok.rc from
 * GConf and loads it.
 */
-void gok_main_read_rc ()
+static void
+gok_main_read_rc ()
 {
     GConfClient *gconf_client = NULL;
     GError *gconf_err = NULL;
@@ -1310,7 +1349,7 @@ void gok_main_read_keyboards ()
     GError *gconf_err = NULL;
     gchar *directory_name = NULL;
     gboolean gconf_keyboard_directory_error = FALSE;
-
+	
     gconf_client = gconf_client_get_default ();
     
     directory_name = gconf_client_get_string (gconf_client,
@@ -1346,9 +1385,10 @@ void gok_main_read_keyboards ()
 
 /**
 * gok_main_read_keyboards_from_dir:
-* @directory: The name of the directory to read the keyboard files from.
+* @directory: The name of the keyboard root directory location.
 * 
-* Reads all the keyboard files from the given directory.
+* Reads all the keyboard files from the given directory or a subdirectory that
+* is a closer match based on user's language if it exists.
 *
 * Returns: A pointer to the first keyboard, NULL if no keyboards could be read.
 */
@@ -1361,16 +1401,40 @@ GokKeyboard* gok_main_read_keyboards_from_dir (const char *directory)
 	struct dirent* pDirectoryEntry;
 	gchar* filename;
 	gchar* complete_path;
+	GList* langlist;
+	gchar dirtemp[1024]; /* no reliable way to avoid magic # here */
 
 	gok_log_enter();
-	gok_log ("directory: %s", directory);
 	pKeyboardFirst = NULL;
 	pKeyboard = NULL;
 	pKeyboardPrevious = NULL;
+	langlist == NULL;
+	pDirectoryKeyboards = NULL;
+
+	/* find the locale specific keyboards */
+	langlist = (GList*) gnome_i18n_get_language_list(NULL);
+	if (langlist != NULL) {
+		gint i = 0;
+		guint l = g_list_length(langlist);
+		while ((pDirectoryKeyboards == NULL) && (i < l)) {
+			strncpy(dirtemp, directory, 950);
+			strncat(dirtemp, G_DIR_SEPARATOR_S, 1);
+			strncat(dirtemp, g_list_nth_data(langlist, i),70);
+			pDirectoryKeyboards = opendir (dirtemp);
+			i++;
+		}
+		g_list_free(langlist);
+	}
+
+	if (pDirectoryKeyboards == NULL) { /* no locale specific directory found */
+		pDirectoryKeyboards = opendir (directory); /* use default */
+	}
+	else {
+		directory = dirtemp; /* used later */
+	}
+		
 
 	/* get a list of all the files from the given directory name */
-	/* first, open the keyboard directory */
-	pDirectoryKeyboards = opendir (directory);
 	if (pDirectoryKeyboards == NULL)
 	{
 		gok_log_x ("Can't open keyboard directory in gok_main_read_files!");
@@ -1384,6 +1448,8 @@ GokKeyboard* gok_main_read_keyboards_from_dir (const char *directory)
 		/* pKeyboard = pKeyboardFirst = gok_keyboard_get_alpha_by_frequency (); *//* TESTING ONLY!!!! */
 		if (pKeyboard)
 			pKeyboard->pKeyboardPrevious = NULL;
+		else
+		        gok_data_set_use_xkb_kbd (FALSE);
 		pKeyboardPrevious = pKeyboard;
 	}
 
@@ -1432,7 +1498,8 @@ GokKeyboard* gok_main_read_keyboards_from_dir (const char *directory)
 * Then calls gok_scanner_initialize passing the directory it
 * got from GConf.
 */
-void gok_main_initialize_access_methods ()
+static void
+gok_main_initialize_access_methods ()
 {
     GConfClient *gconf_client = NULL;
     GError *gconf_err = NULL;
@@ -1472,7 +1539,8 @@ void gok_main_initialize_access_methods ()
 * Retreives the name of the directory containing dictionary.txt from
 * GConf and calls gok_wordcomplete_open passing that directory.
 */
-void gok_main_initialize_wordcomplete ()
+static void
+gok_main_initialize_wordcomplete ()
 {
     GConfClient *gconf_client = NULL;
     GError *gconf_err = NULL;
@@ -1510,7 +1578,8 @@ void gok_main_initialize_wordcomplete ()
 * Retreives the name of the directory containing dictionary.txt from
 * GConf and calls gok_wordcomplete_open passing that directory.
 */
-void gok_main_initialize_commandprediction ()
+static void
+gok_main_initialize_commandprediction ()
 {
 	gok_log_enter();
     /* intialize command predictor */
@@ -1816,7 +1885,8 @@ void gok_main_store_window_center ()
 * Delete any keyboards that were created.
 * This must be called at the end of the program.
 */
-void gok_main_close()
+void
+gok_main_close()
 {
 	/* NOTE: the order in which things are cleaned up is very important */
 
@@ -1825,6 +1895,7 @@ void gok_main_close()
 
 	/* unhook listeners/callbacks */
 	gok_spy_deregister_windowchangelistener ((void *)gok_main_window_change_listener);
+	gok_spy_deregister_objectstatelistener ((void *)gok_main_object_state_listener);
 	/* the call to deregister_mousebuttonlistener is in on_window1_destroy */
 	
 	gok_spy_stop();
@@ -1907,7 +1978,8 @@ void gok_main_display_gconf_error ()
 * 
 * Displays a dialog informing user that the given geometry was not correct.
 */
-void gok_main_display_geometry_error ()
+static void
+gok_main_display_geometry_error ()
 {
 	GtkWidget* dialog;
         
@@ -2021,7 +2093,7 @@ gok_main_get_login (void)
 * 
 * returns: gboolean
 */
-gboolean
+static gboolean
 gok_main_has_xkb_extension ()
 {
 	/* TODO/revisit - is this a good check? */
@@ -2032,7 +2104,7 @@ gok_main_has_xkb_extension ()
 	return FALSE;
 }
 
-gboolean
+static gboolean
 gok_main_check_sticky_keys (GtkWidget *widget)
 {
 	int op_rtn, event_rtn, error_rtn;
@@ -2066,7 +2138,7 @@ gok_main_check_sticky_keys (GtkWidget *widget)
 #endif
 				xkb_ok = TRUE;
 				if (!bLoginOption) {
-				  dialog = gtk_message_dialog_new (GTK_WINDOW (widget),
+				  dialog = gtk_message_dialog_new (NULL,
 								 GTK_DIALOG_DESTROY_WITH_PARENT,
 								 GTK_MESSAGE_INFO,
 								 GTK_BUTTONS_OK,
@@ -2087,7 +2159,7 @@ gok_main_check_sticky_keys (GtkWidget *widget)
 	if (xkb_ok == FALSE) {
 		if (!bLoginOption) {
                 /* post an error dialog */
-			dialog = gtk_message_dialog_new (GTK_WINDOW (widget),
+			dialog = gtk_message_dialog_new (NULL,
 						 GTK_DIALOG_DESTROY_WITH_PARENT,
 						 GTK_MESSAGE_ERROR,
 						 GTK_BUTTONS_CLOSE,
@@ -2108,13 +2180,20 @@ static GtkWidget* acd = NULL; /* accessibility check dialog */
 
 void check_accessibility_cb ( GObject* o, gpointer* data )
 {
+	GnomeClient *client;
+
 	gtk_widget_destroy(acd);	
 	
 	if (strcmp((gchar*)data,"logout") == 0) {
 		gok_gconf_set_bool ( gok_data_get_gconf_client(),
 			GCONF_ACCESSIBILITY_KEY, TRUE );
-			gok_main_close();  /* FIXME: need to logout */
-			_exit (0);
+		if (!(client = gnome_master_client())) {
+			gok_main_close();
+		}
+ 		gnome_client_request_save (client, GNOME_SAVE_GLOBAL, TRUE,
+			GNOME_INTERACT_ANY, FALSE, TRUE); /* code borrowed from 
+			gnome-control-center/ capplets/ accessibility/ at-properties/ 
+			main.c */
 	}
 	else if (strcmp((gchar*)data,"quit") == 0) {
 		gok_main_close();
@@ -2143,9 +2222,9 @@ gok_main_check_accessibility ()
 		gboolean 	returnCode = TRUE;
 		
 		acd = gtk_message_dialog_new (
-			GTK_WINDOW(m_pWindowMain), 
+			NULL, 
 			GTK_DIALOG_DESTROY_WITH_PARENT,
-			GTK_MESSAGE_WARNING,
+			GTK_MESSAGE_INFO, /* FIXME: GTK_MESSAGE_WARNING occludes keyboard */
 			GTK_BUTTONS_NONE,			
 			_("Assistive Technology Support Is Not Enabled." 
 		"\n\n" 
